package de.desy.web2cToolkit.web2c.web2cComponents;

import java.awt.Color;
import java.awt.Font;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Random;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.AxisLocation;
import org.jfree.chart.axis.LogarithmicAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.StandardXYItemRenderer;
import org.jfree.chart.renderer.xy.XYAreaRenderer;
import org.jfree.data.time.Millisecond;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;
import org.jfree.data.xy.XYDataset;
import de.desy.web2cToolkit.web2c.web2cChatUtilities.Web2cChatChannel;
import de.desy.web2cToolkit.web2c.web2cDeviceAccess.Web2cCommonDeviceAccess;
import de.desy.web2cToolkit.web2c.web2cVideoUtilities.Web2cCommonVideoConnector;

/**
 * Trend-type Plot
 * @author R.Bacher, DESY
 */
public class Web2cTrend extends Web2cComponent {
	
	private Random randomGenerator = null;
	private String web2cSession = null;
	private float[][] dataValues = null;
	private int[][] dataTimeStamps = null;
	private int numberOfDataPoints = 0;
	private int numberOfDataTraces = 0;
	private String[] dataAddressName = null;
	private String title = null;
	private int dataArrayCompletionLength = 0;	
	private String fromInSeconds = null;
	private String toInSeconds= null;
	
	/**
	 * Trend-type Plot
	 * @param refId ID of reference component which component belongs to
	 * @param home directory which contains user-specific files
	 * @param lang language
	 * @param x horizontal scaling factor
	 * @param y vertical scaling factor
	 * @param names list of valid attribute names
	 * @param values list of actual attribute values
	 * @param simulatedData flag for simulated or real data
	 * @param privilege actual user privilege
	 * @param cda actual common device access instance
	 * @param cvc actual common video connector instance
	 * @param ch actual chat channel instance
	 * @param session Web2c session id
	 */
	public Web2cTrend(String refId, String home, int lang, double x, double y, ArrayList<String> names, ArrayList<String> values, boolean simulatedData, String privilege, Web2cCommonDeviceAccess cda, Web2cCommonVideoConnector cvc, Web2cChatChannel ch, String session) {

		super(refId, home, lang, x, y, names, values, simulatedData, privilege, cda, cvc, ch);
		
		if (extensionManifest.equals("")) componentDataSize = property.size();
		registerMonitors(false);
		initSoundAndBlinking(componentDataSize);

		web2cSession = session;
		randomGenerator = new Random();		
		
		initComponent();
	}
	
	/**
	 * Initializes the component
	 */
	public void initComponent() {
		
		int i = 0;
		int j = 0;
		int now = 0;
		int from = 0;
		int to = 0;
		Calendar calendar = null;
		
		calendar = Calendar.getInstance();
		calendar.setTime(new Date());
		now = (int)(calendar.getTimeInMillis() / 1000.0);
		fromInSeconds = String.valueOf(now - Integer.valueOf(timeIntervalInSeconds));
		toInSeconds = String.valueOf(now);		
		from = Integer.valueOf(fromInSeconds);
		to = Integer.valueOf(toInSeconds);		
		numberOfDataPoints = to - from + 1;
		
		dataArrayCompletionLength = numberOfDataPoints;
		numberOfDataTraces = property.size();
		dataAddressName = new String[numberOfDataTraces];
		dataTimeStamps = new int[numberOfDataTraces][dataArrayCompletionLength];
		dataValues = new float[numberOfDataTraces][dataArrayCompletionLength];
		for (j = 0; j < numberOfDataTraces; j++) {
			for (i = 0; i < dataArrayCompletionLength; i++) {
				dataTimeStamps[j][i] = i + from;
				dataValues[j][i] = 0;
			}
			dataAddressName[j] = "";
		}
		title = "";
	}

	/**
	 * Returns component's HTML tags 
	 * @return HTML tags
	 */
	public ArrayList<String> getHTML() {
		
		ArrayList<String> componentHTML = new ArrayList<String>();	

		if (userPrivilege.equals(layman) && expertComponent.equals(yes)) {
			componentHTML.add("");
		}
		else {	
			if (propertyLinksEstablished == true) {
				componentHTML.add("			<img id=\"" + id + "\" src=\"Web2c?param=<plot><sessionId>" + web2cSession + "</sessionId><id>" + id + "</id><timeStamp>" + toInSeconds + "</timeStamp></plot>\" title=\"" + toolTipText + "\" alt=\"Image temporarily unavailable\" style=\"position:absolute; left:" + marginLeft + "; top:" + marginTop + "; width:" + width + "; height:" + height + "\">");
			}
			else {
				componentHTML.add("");
			}
		}
		
		return componentHTML;
	}

	/**
	 * Returns component's JavaScript code snippets
	 * @return JavaScript code snippets
	 */
	public ArrayList<String> getJavaScript() {
		
		ArrayList<String> componentJavaScript = new ArrayList<String>();	
		
		if (userPrivilege.equals(layman) && expertComponent.equals(yes)) {
			componentJavaScript.add("");
		}
		else {	
			if (propertyLinksEstablished == true) {
				componentJavaScript.add("	//Trend: " + id);
			}
			else {
				componentJavaScript.add("");
			}
		}

		return componentJavaScript;
	}
	
	/**
	 * Returns simulated data
	 * @return data (xml-type)
	 */
	public String getSimulatedData() {
		
		String str = "";
		String identifier = "";
		String value = "";
		String plot = "";
		Calendar calendar = null;
		int now = 0;
		int from = 0;
		int to = 0;
		int i = 0;
		int j = 0;

		calendar = Calendar.getInstance();
		calendar.setTime(new Date());
		now = (int)(calendar.getTimeInMillis() / 1000.0);
		fromInSeconds = String.valueOf(now - Long.valueOf(timeIntervalInSeconds));
		toInSeconds = String.valueOf(now);		
		from = Integer.valueOf(fromInSeconds);
		to = Integer.valueOf(toInSeconds);		
		numberOfDataPoints = to - from + 1;		
		
		title = caption.get(0);
		plot = "<plot><sessionId>" + web2cSession + "</sessionId><id>" + id + "</id><timeStamp>" + toInSeconds + "</timeStamp></plot>";
		
		dataArrayCompletionLength = numberOfDataPoints;
		for (j = 0; j < numberOfDataTraces; j++) {
			for (i = 0; i < dataArrayCompletionLength; i++) {
				dataTimeStamps[j][i] = i + from;
			}
			for (i = 0; i < dataArrayCompletionLength - 1; i++) {
				dataValues[j][i] = dataValues[j][i + 1];
			}
			dataValues[j][dataArrayCompletionLength - 1] = randomGenerator.nextInt(100);
			dataAddressName[j] = "Simulated Data " + String.valueOf(j);
		}
		
		identifier = "<id>" + id + "</id>";
		value = "<value>" + plot + "</value>";

		str = "<update>" + identifier + value + "</update>";

		return str;
	}
	
	/**
	 * Returns real data from common device access plug or from executing the extension code
	 * @return data (xml-type)
	 */
	public String getRealData() {
		
		String str = "";
		String identifier = "";
		String value = "";
		String style = "";
		String plot = "";
		String[] ret = new String[componentDataSize];
		String[] act = null;
		Calendar calendar = null;
		int now = 0;
		int from = 0;
		int to = 0;
		int i = 0;
		int j = 0;

		calendar = Calendar.getInstance();
		calendar.setTime(new Date());
		now = (int)(calendar.getTimeInMillis() / 1000.0);
		fromInSeconds = String.valueOf(now - Long.valueOf(timeIntervalInSeconds));
		toInSeconds = String.valueOf(now);		
		from = Integer.valueOf(fromInSeconds);
		to = Integer.valueOf(toInSeconds);		
		numberOfDataPoints = to - from + 1;
		
		title = caption.get(0);
		plot = "<plot><sessionId>" + web2cSession + "</sessionId><id>" + id + "</id><timeStamp>" + toInSeconds + "</timeStamp></plot>";

		if (!extensionManifest.equals("")) {
			if (extensionPropertyLinksEstablished == true) {
				extensionExecutor.executeExtensionCode("live");
				for (i = 0; i < ret.length; i++) {
					ret[i] = (String)(extensionExecutor.getData()[i]);
					ret[i] = ret[i].split(">")[1].split("<")[0];	
					dataAddressName[i] = "Trace" + i;
				}
			}
			if (!userPrivilege.equals(layman) && extensionExecutor.getActions() != null) {
				act = extensionExecutor.getActions();
				for (i = 0; i < act.length; i++) {
					commonDeviceAccess.execute(act[i], true);
				}
			}
		}
		else {
			if (propertyLinksEstablished == true) {
				for (i = 0; i < ret.length; i++) {
					String address = (property.get(i));
					ret[i] = (commonDeviceAccess.getMonitorData(address)).split(">")[1].split("<")[0];	
					dataAddressName[i] = address;
				}
			}
		}
		
		if (extensionPropertyLinksEstablished == true || propertyLinksEstablished == true) {
			dataArrayCompletionLength = numberOfDataPoints;
			for (j = 0; j < numberOfDataTraces; j++) {
				for (i = 0; i < dataArrayCompletionLength; i++) {
					dataTimeStamps[j][i] = i + from;
				}
				for (i = 0; i < dataArrayCompletionLength - 1; i++) {
					dataValues[j][i] = dataValues[j][i + 1];
				}
				dataValues[j][dataArrayCompletionLength - 1] = Float.valueOf(characterEscaper.decodeString(ret[j]));
				
				for (i = 0; i < dataArrayCompletionLength; i++) {
					if (functionType.contains("power")) {
						if (functionParameter.equals("") || functionParameter.equals(" ")) {
							functionParameter = "1.0";
						}
						dataValues[j][i] = (float)(Math.pow((double)dataValues[j][i], Double.valueOf(functionParameter)));
					}
					if (functionType.contains("factor")) {
						if (functionParameter.equals("") || functionParameter.equals(" ")) {
							functionParameter = "1.0";
						}
						dataValues[j][i] =  dataValues[j][i] * Float.valueOf(functionParameter);
					}
					if (functionType.contains("offset")) {
						if (functionParameter.equals("") || functionParameter.equals(" ")) {
							functionParameter = "0.0";
						}
						dataValues[j][i] =  dataValues[j][i] + Float.valueOf(functionParameter);
					}
					if (functionType.contains("log")) {
						if (functionParameter.equals("") || functionParameter.equals(" ")) {
							functionParameter = "1.0";
						}
						if (dataValues[j][i] == 0.0) {
							 dataValues[j][i] = 1.0f;
						}
						dataValues[j][i] = dataValues[j][i] * Float.valueOf(functionParameter);
					}
				}
			}
			
			if (functionType.contains("norm")) {
				if (functionParameter.equals("") || functionParameter.equals(" ")) {
					functionParameter = "1";
				}
				for (j = 0; j < numberOfDataTraces; j++) {
					if (j != (Integer.valueOf(functionParameter) - 1)) {
						for (i = 0; i < dataArrayCompletionLength; i++) {
							dataValues[j][i] =  (dataValues[j][i] / dataValues[Integer.valueOf(functionParameter) - 1][i]) * 100.0f;
						}
					}
				}
				for (i = 0; i < dataArrayCompletionLength; i++) {
					dataValues[Integer.valueOf(functionParameter) - 1][i] = 100.0f;
				}
			}
			if (functionType.contains("ref")) {
				if (functionParameter.equals("") || functionParameter.equals(" ")) {
					functionParameter = "1";
				}
				for (j = 0; j < numberOfDataTraces; j++) {
					if (j != (Integer.valueOf(functionParameter) - 1)) {
						for (i = 0; i < dataArrayCompletionLength; i++) {
							dataValues[j][i] =  dataValues[j][i] - dataValues[Integer.valueOf(functionParameter) - 1][i];
						}
					}
				}
				for (i = 0; i < dataArrayCompletionLength; i++) {
					dataValues[Integer.valueOf(functionParameter) - 1][i] = 0.0f;
				}
			}
			
			identifier = "<id>" + id + "</id>";
			value = "<value>" + plot + "</value>";
			style = getStyleUpdate();
	
			str = "<update>" + identifier + value + style + "</update>";
		}
		
		return str;
	}  	
	
	/**
	 * Returns chart width [px]
	 * @return width of chart
	 */
	public int getChartWidth() {
		
		return Integer.valueOf(width.substring(0,width.length() - 2));		
	}
	
	/**
	 * Returns chart height [px]
	 * @return height of chart
	 */
	public int getChartHeight() {
		
		return Integer.valueOf(height.substring(0,height.length() - 2));		
	}
	
	/**
	 * Returns chart
	 * @return chart
	 */
	public JFreeChart getChart() {
		
		int i = 0;
		JFreeChart chart = null;

		try {
			chart = ChartFactory.createTimeSeriesChart(title , "Date", "", null, true, true, false);
			XYPlot plot = chart.getXYPlot();	
			ValueAxis domainAxis = plot.getDomainAxis();
			domainAxis.setLabelFont(new Font("SansSerif", Font.BOLD, Integer.valueOf(fontSize.substring(0, fontSize.length() - 2))));
			domainAxis.setTickLabelFont(new Font("SansSerif", Font.BOLD, Integer.valueOf(fontSize.substring(0, fontSize.length() - 2))));
			plot.setDomainAxis(domainAxis);
			plot.setForegroundAlpha(1.0f);
			if (plotAxis.equals(multi)) {
				for (i = 0; (i < numberOfDataTraces) && (i < maxTraces); i++) {
					if (functionType.contains("log")) {
						LogarithmicAxis axis = new LogarithmicAxis((String)(legend.toArray()[i]));
						axis.setLog10TickLabelsFlag(true);
						axis.setAllowNegativesFlag(true);
						if (!((Double)(yMin.toArray()[0])).equals((Double)(yMax.toArray()[0]))) {
							axis.setAutoRange(false);
							axis.setRange((Double)(yMin.toArray()[i]), (Double)(yMax.toArray()[i]));
						}
						else {
							axis.setAutoRange(true);
						}
						axis.setLabelPaint(plotColor[i]);
						axis.setTickLabelPaint(plotColor[i]);	
						axis.setLabelFont(new Font("SansSerif", Font.BOLD, Integer.valueOf(fontSize.substring(0, fontSize.length() - 2))));	
						axis.setTickLabelFont(new Font("SansSerif", Font.BOLD, Integer.valueOf(fontSize.substring(0, fontSize.length() - 2))));
						plot.setRangeAxis(i,axis);
						plot.setRangeAxisLocation(i, AxisLocation.BOTTOM_OR_LEFT);
					}
					else {
						NumberAxis axis = new NumberAxis((String)(legend.toArray()[i]));
						if (!((Double)(yMin.toArray()[0])).equals((Double)(yMax.toArray()[0]))) {
							axis.setAutoRange(false);
							axis.setRange((Double)(yMin.toArray()[i]), (Double)(yMax.toArray()[i]));
						}
						else {
							axis.setAutoRange(true);
						}
						axis.setLabelPaint(plotColor[i]);
						axis.setTickLabelPaint(plotColor[i]);	
						axis.setLabelFont(new Font("SansSerif", Font.BOLD, Integer.valueOf(fontSize.substring(0, fontSize.length() - 2))));	
						axis.setTickLabelFont(new Font("SansSerif", Font.BOLD, Integer.valueOf(fontSize.substring(0, fontSize.length() - 2))));
						plot.setRangeAxis(i,axis);
						plot.setRangeAxisLocation(i, AxisLocation.BOTTOM_OR_LEFT);
					}
					XYDataset dataset = createDataset(i);
					plot.setDataset(i, dataset);
					plot.mapDatasetToRangeAxis(i, i);
					if (plotType.equals(linePlot)) {
						StandardXYItemRenderer renderer = new StandardXYItemRenderer();
						renderer.setSeriesPaint(0, plotColor[i]);
						plot.setRenderer(i, renderer);
					}
					else if (plotType.equals(areaPlot)) {
						XYAreaRenderer renderer = new XYAreaRenderer();
						renderer.setSeriesPaint(0, plotColor[i]);
						plot.setRenderer(i, renderer);
					}
				}
			}
			else if (plotAxis.equals(single)){
				if (functionType.contains("log")) {
					LogarithmicAxis axis = new LogarithmicAxis("");
					axis.setLog10TickLabelsFlag(true);
					axis.setAllowNegativesFlag(true);
					if (!((Double)(yMin.toArray()[0])).equals((Double)(yMax.toArray()[0]))) {
						axis.setAutoRange(false);
						axis.setRange((Double)(yMin.toArray()[0]), (Double)(yMax.toArray()[0]));
					}
					else {
						axis.setAutoRange(true);
					}
					axis.setLabelPaint(Color.BLACK);
					axis.setTickLabelPaint(Color.BLACK);	
					axis.setLabelFont(new Font("SansSerif", Font.BOLD, Integer.valueOf(fontSize.substring(0, fontSize.length() - 2))));	
					axis.setTickLabelFont(new Font("SansSerif", Font.BOLD, Integer.valueOf(fontSize.substring(0, fontSize.length() - 2))));
					plot.setRangeAxis(0,axis);
					plot.setRangeAxisLocation(0, AxisLocation.BOTTOM_OR_LEFT);
				}
				else {
					NumberAxis axis = new NumberAxis("");
					if (!((Double)(yMin.toArray()[0])).equals((Double)(yMax.toArray()[0]))) {
						axis.setAutoRange(false);
						axis.setRange((Double)(yMin.toArray()[0]), (Double)(yMax.toArray()[0]));
					}
					else {
						axis.setAutoRange(true);
					}
					axis.setLabelPaint(Color.BLACK);
					axis.setTickLabelPaint(Color.BLACK);	
					axis.setLabelFont(new Font("SansSerif", Font.BOLD, Integer.valueOf(fontSize.substring(0, fontSize.length() - 2))));	
					axis.setTickLabelFont(new Font("SansSerif", Font.BOLD, Integer.valueOf(fontSize.substring(0, fontSize.length() - 2))));
					plot.setRangeAxis(0,axis);
					plot.setRangeAxisLocation(0, AxisLocation.BOTTOM_OR_LEFT);
				}
				for (i = 0; (i < numberOfDataTraces) && (i < maxTraces); i++) {
					XYDataset dataset = createDataset(i);
					plot.setDataset(i, dataset);
					if (plotType.equals(linePlot)) {
						StandardXYItemRenderer renderer = new StandardXYItemRenderer();
						renderer.setSeriesPaint(0, plotColor[i]);
						plot.setRenderer(i, renderer);
					}
					else if (plotType.equals(areaPlot)) {
						XYAreaRenderer renderer = new XYAreaRenderer();
						renderer.setSeriesPaint(0, plotColor[i]);
						plot.setRenderer(i, renderer);
					}
				}
			}
		}
		catch (Exception e) {
			System.out.println(e.toString());
			e.printStackTrace();
			System.out.println("Error generating chart!");
		}
		
		return chart;
	}
	
	/**
	 * Returns data set
	 * @param i trace index
	 * @return data set
	 */
	private XYDataset createDataset(int i) {
		
		int j = 0;
		Calendar calendar = null;
		
		TimeSeries series = new TimeSeries((String)(legend.toArray()[i]), Millisecond.class);	
		calendar = Calendar.getInstance();	

		for (j = 0; j < dataArrayCompletionLength; j++) {
			if (dataTimeStamps[i][j] != 0) {
				calendar.setTimeInMillis((long)(dataTimeStamps[i][j] * 1000.0));
				if (functionType.contains("log") && (dataValues[i][j] <= 0)) dataValues[i][j] = 1;
				series.addOrUpdate(new Millisecond(calendar.getTime()),(double)(dataValues[i][j]));
			}
		}
		
		TimeSeriesCollection dataset = new TimeSeriesCollection();
		dataset.addSeries(series);
		
		return dataset;
	}
}
