package de.desy.web2cToolkit.web2c.web2cComponents;

import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Calendar;
import de.desy.web2cToolkit.web2c.web2cChatUtilities.Web2cChatChannel;
import de.desy.web2cToolkit.web2c.web2cDeviceAccess.Web2cCommonDeviceAccess;
import de.desy.web2cToolkit.web2c.web2cVideoUtilities.Web2cCommonVideoConnector;
import de.desy.web2cToolkit.web2c.web2cVideoUtilities.Web2cRTPVideoReceiver;
import de.desy.web2cToolkit.web2c.web2cVideoUtilities.Web2cTINEVideoReceiver;

/**
 * Video Image
 * @author R.Bacher, DESY
 */
public class Web2cVideo extends Web2cComponent {
	
	private String nowTime = null;
	private String web2cSession = null;
	private Web2cRTPVideoReceiver rtpClient = null;
	private Web2cTINEVideoReceiver tineClient = null;
	
	/**
	 * Video Image
	 * @param refId ID of reference component which component belongs to
	 * @param home directory which contains user-specific files
	 * @param lang language
	 * @param x horizontal scaling factor
	 * @param y vertical scaling factor
	 * @param names list of valid attribute names
	 * @param values list of actual attribute values
	 * @param simulatedData flag for simulated or real data
	 * @param privilege actual user privilege
	 * @param cda actual common device access instance
	 * @param cvc actual common video connector instance
	 * @param ch actual chat channel instance
	 * @param session Web2c session id
	 */
	public Web2cVideo(String refId, String home, int lang, double x, double y, ArrayList<String> names, ArrayList<String> values, boolean simulatedData, String privilege, Web2cCommonDeviceAccess cda, Web2cCommonVideoConnector cvc, Web2cChatChannel ch, String session) {

		super(refId, home, lang, x, y, names, values, simulatedData, privilege, cda, cvc, ch);
		
		componentDataSize = 1;
		registerReceivers();
		initSoundAndBlinking(1);
		
		web2cSession = session;	
		
		initComponent();
	}
	
	/**
	 * Initializes the component
	 */
	public void initComponent() {
		
		if (borderStyle.equals(solid)) {
			width = String.valueOf(Integer.valueOf(width.split("p")[0]) - (2 * Integer.valueOf(borderWidth.split("p")[0]))) + "px";
			height = String.valueOf(Integer.valueOf(height.split("p")[0]) - (2 * Integer.valueOf(borderWidth.split("p")[0]))) + "px";
		}
	}

	/**
	 * Returns the video client attached to this component
	 * @return video client
	 */
	public Web2cRTPVideoReceiver getRTPClient() {
		
		return rtpClient;
	}
	
	/**
	 * Returns the tine client attached to this component
	 * @return tine client
	 */
	public Web2cTINEVideoReceiver getTINEClient() {
		
		return tineClient;
	}

	/**
	 * Returns component's HTML tags
	 * @return HTML tags
	 */
	public ArrayList<String> getHTML() {
		
		ArrayList<String> componentHTML = new ArrayList<String>();	
		String html = null;

		if (userPrivilege.equals(layman) && expertComponent.equals(yes)) {
			componentHTML.add("");
		}
		else {	
			html = "			<div id=\"" + id + "\" title=\"" + toolTipText + "\"";
			html = html + " style=\"position:absolute; left:" + marginLeft + "; top:" + marginTop + "; width:" + width + "; height:" + height + "; color:" + color + "; background-color:" + backgroundColor + "; border-style:" + borderStyle + "; border-width:" + borderWidth + "; border-color:" + borderColor + "; font-size:" + fontSize + "; font-weight:" + fontThickness + "; text-align:" + textAlignment + "\">";	
			componentHTML.add(html);
			
			if (caption.get(0).equals("") || caption.get(0).equals(" ")) {
			}
			else {
				int headerCellWidth = (Integer.valueOf(width.split("p")[0])) - (2 * Integer.valueOf(borderWidth.split("p")[0]));
				int headerCellHeight = Math.round(0.1f * ((Integer.valueOf(height.split("p")[0])) - (2 * Integer.valueOf(borderWidth.split("p")[0]))));
				int headerCellXPos = 0;
				int headerCellYPos = 0;
				html = "			   <div id=\"" + id + "_header" + "\" title=\"" + toolTipText + "\"";
				html = html + " style=\"position:absolute; left:" + ((Integer)(headerCellXPos)).toString() + "px" + "; top:" + ((Integer)(headerCellYPos)).toString() + "px" + "; width:" + ((Integer)(headerCellWidth)).toString() + "px" + "; height:" + ((Integer)(headerCellHeight)).toString() + "px" + "; color:" + color + "; background-color:" + backgroundColor + "; border-style:" + "none" + "; border-width:" + borderWidth + "; border-color:" + borderColor + "; font-size:" + fontSize + "; font-weight:" + fontThickness + "; text-align:" + textAlignment + "\">" + caption.get(0) + "</div>";	
				componentHTML.add(html);
				
				int videoCellWidth = (Integer.valueOf(width.split("p")[0]));
				int videoCellHeight = Math.round(0.9f * ((Integer.valueOf(height.split("p")[0]))));
				int videoCellXPos = 0;
				int videoCellYPos = headerCellHeight  + (0 * Integer.valueOf(borderWidth.split("p")[0]));
				html = "			   <img id=\"" + id + "_video" + "\" src=\"Web2c?param=<vimg><sessionId>" + web2cSession + "</sessionId><id>" + id + "</id><timeStamp>" + nowTime + "</timeStamp></vimg>\" title=\"" + toolTipText + "\" alt=\"Image temporarily unavailable\"";
				html = html + " style=\"position:absolute; left:" + ((Integer)(videoCellXPos)).toString() + "px" + "; top:" + ((Integer)(videoCellYPos)).toString() + "px" + "; width:" + ((Integer)(videoCellWidth)).toString() + "px" + "; height:" + ((Integer)(videoCellHeight)).toString() + "px" + "; border-style:" + "none" + "; border-width:" + borderWidth + "; border-color:" + borderColor + "\">";
				componentHTML.add(html);
				
			}
			html = "			</div>";
			componentHTML.add(html);
		}
		
		return componentHTML;
	}

	/**
	 * Returns component's JavaScript code snippets
	 * @return JavaScript code snippets
	 */
	public ArrayList<String> getJavaScript() {
		
		ArrayList<String> componentJavaScript = new ArrayList<String>();	
		
		if (userPrivilege.equals(layman) && expertComponent.equals(yes)) {
			componentJavaScript.add("");
		}
		else {	
			componentJavaScript.add("	//Video: " + id);
		}

		return componentJavaScript;
	}
	
	/**
	 * Returns simulated data
	 * @return data (xml-type)
	 */
	public String getSimulatedData() {
		
		String str = "";
		String identifier = "";
		String value = "";
		String videoSimulation = web2cdocs + directorySeparator + simulatedVideoImage;
		
		identifier = "<id>" + id + "_video" + "</id>";
		value = "<value>" + videoSimulation + "</value>";
		
		str = "<update>" + identifier + value + "</update>";
		
		return str;
	}
	
	/**
	 * Returns real data from common device access plug
	 * @return data (xml-type)
	 */
	public String getRealData() {
		
		String str = "";
		String identifier = "";
		String value = "";
		String style = "";
		String blinking = "";
		Calendar calendar = null;
		int now = 0;
		
		calendar = Calendar.getInstance();
		now = (int)(calendar.getTimeInMillis() / 1000.0);
		nowTime = String.valueOf(now);	
		
		identifier = "<id>" + id + "_video" + "</id>";
		value = "<value>" + "<vimg><sessionId>" + web2cSession + "</sessionId><id>" + id + "</id><timeStamp>" + nowTime + "</timeStamp></vimg>" + "</value>";
		style = getStyleUpdate();
		blinking = getBlinking(0);
		
		str = "<update>" + identifier + value + style + blinking + "</update>";
		
		return str;
	} 
	
	/**
 	 * Returns the video stream
 	 * @param frame output stream
	 */
	public void getVideoFrame(ByteArrayOutputStream frame) {
		
		String[] act = null;
		
		if (!extensionManifest.equals("")) {
			extensionExecutor.executeExtensionCode("");
			frame = (ByteArrayOutputStream)(extensionExecutor.getData()[0]);
			if (!userPrivilege.equals(layman) && extensionExecutor.getActions() != null) {
				act = extensionExecutor.getActions();
				for (int i = 0; i < act.length; i++) {
					commonDeviceAccess.execute(act[i], true);
				}
			}
		}
		else {
			commonVideoConnector.getVideo(frame, property.get(0));
		}
	}
}
