package de.desy.web2cToolkit.web2cGateway;

import java.io.File;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.Observable;
import java.util.Observer;
import de.desy.web2cToolkit.web2c.web2cDeviceAccess.Web2cCommonDeviceAccess;
import de.desy.web2cToolkit.web2cUtil.web2cDebug.Web2cFileLogger;
import de.desy.web2cToolkit.web2cUtil.web2cDebug.Web2cKeyBoardListener;
import de.desy.web2cToolkit.web2cUtil.web2cEscape.Web2cASCIICharacterEscaper;
import de.desy.web2cToolkit.web2cUtil.web2cFileIO.Web2cConfigReader;
import de.desy.web2cToolkit.web2cUtil.web2cInterfaces.IWeb2cConstants;

/**
 * Page
 * @author R.Bacher, DESY
 */
public class Web2cGatewayPage implements IWeb2cConstants, Observer {

	private String homeDir = null;
	private String userName = null;
	private String userAddress = null;
	private String userPrivilege = null;
	private String networkPrivilege = null;
	private Web2cCommonDeviceAccess commonDeviceAccess = null;
	private Web2cFileLogger debugLogger = null;
	private Web2cKeyBoardListener debugListener = null;
	private String web2cSession = null;
	private String pageConfigFile = null;
	private String error = null;
	private ArrayList<String> monitorList = new ArrayList<String>();
	private ArrayList<MonitorLinkUpdater> monitorLinkUpdaterList = new ArrayList<MonitorLinkUpdater>();
	private Web2cASCIICharacterEscaper characterEscaper = null;

	private boolean exclusive = false;
	private String registerStatus = "idle";
	private String registerError = "";
	

	/**
	 * Page
	 * @param session Web2c session id
	 * @param home directory which contains user-specific files
	 * @param configFile actual configuration file
	 * @param name actual user name
	 * @param ipAddress actual IP address of user
	 * @param cda actual common device access instance
	 * @param logger debug logger
	 * @param listener debug flag listener
	 */
	public Web2cGatewayPage(String session, String home, String configFile, String name, String ipAddress, Web2cCommonDeviceAccess cda, Web2cFileLogger logger, Web2cKeyBoardListener listener) {

		web2cSession = session;	
		homeDir = home;
		userPrivilege = layman;
		networkPrivilege = layman;
		userName = "guest";
		userAddress = "";
		if (name == null) {
			userName = "guest";
		}
		else {
			userName = name;
		}
		if (ipAddress == null) {
			userAddress = "";
		}
		else {
			userAddress = ipAddress;
		}
		debugLogger = logger;
		debugListener = listener;
		commonDeviceAccess = cda;
		pageConfigFile = configFile;
		registerStatus = "idle";
		registerError = "";
		characterEscaper = new Web2cASCIICharacterEscaper();
		constructPage();
	}
	
	/**
	 * Returns the session id of the page
	 * @return session id
	 */
	public String getSessionId() {
		
		return web2cSession;
	}

	/**
	 * Returns the control system monitors attached to this page
	 * @return list of monitors
	 */
	public ArrayList<String> getMonitors() {
		
		return monitorList;
	}
	
	/**
	 * Terminates updating by gateway
	 */
	public void stopUpdate() {

		synchronized (monitorLinkUpdaterList) {
			for (Iterator<MonitorLinkUpdater> it = monitorLinkUpdaterList.iterator(); it.hasNext();) {
				MonitorLinkUpdater linkUpdater = (MonitorLinkUpdater)(it.next());
				linkUpdater.stopUpdate();
			}
		}
	}
	
	/**
	 * Executes requested action
	 * @param action action to be performed
	 * @return data
	 */
	public String doAction(String action) {
		
		String[] str = null;
		String ret = "";
		
		str = action.split("</do><parameter>");
		str[0] = str[0].replace("<do>", "");
		str[1] = str[1].replace("</parameter>", "");

		if (exclusive == false) {
			if (str.length > 1) {
				if (str[0].equals("update")) ret = doCdaUpdate(str[1]);
				if (str[0].equals("register")) ret = doCdaRegister(str[1]);
				if (str[0].equals("execute")) ret = doCdaExecute(str[1]);
			}
		}
		
		return ret;
	}
	
	/**
	 * Update from the monitor link registerer
	 * @param registerer monitor link registerer
	 * @param status final register status
	 */
	public void update(Observable registerer, Object status) {

		registerError = (String)status;
		registerStatus = "registered";	
	}
	
	/**
	 * Returns data to update client
	 * @param uriList list of control system addresses (Format: &#060;uri&#062;protocol&#058;&#047;&#047;address&#060;&#047;uri&#062;...&#060;uri&#062;protocol&#058;&#047;&#047;address&#060;&#047;uri&#062;)
	 * @return data
	 */
	public String doCdaUpdate(String uriList) {
		
		String[] str = null;	
		ArrayList<String> list = new ArrayList<String>();
		
		str = uriList.split("</uri>");
		for (int i = 0; i < str.length; i++) {
			list.add(str[i].replace("<uri>", ""));
		}

		String update = "";
		synchronized (monitorList) {
			synchronized (monitorLinkUpdaterList) {
				for (Iterator<String> it = monitorList.iterator(); it.hasNext();) {
					String prop = (String)(it.next());
					if (list.contains(prop)) {
						update = update + monitorLinkUpdaterList.get(monitorList.indexOf(prop)).getMostRecentUpdate();
					}
				}
			}
		}
		
		update = "<return>" + update + "<timeStamp>" + String.valueOf(Calendar.getInstance().getTimeInMillis()) + "</timeStamp></return>";
		
		return update;
	}
	
	/**
	 * Registers control system monitors
	 * @param uriList list of control system addresses (Format: &#060;uri&#062;protocol&#058;&#047;&#047;address&#060;&#047;uri&#062;...&#060;uri&#062;protocol&#058;&#047;&#047;address&#060;&#047;uri&#062;)
	 * @return data
	 */
	public String doCdaRegister(String uriList) {

		String[] str = null;
		ArrayList<String> list = new ArrayList<String>();
		
		str = uriList.split("</uri>");
		for (int i = 0; i < str.length; i++) {
			list.add(str[i].replace("<uri>", ""));
		}
		
		error = "";	
		if (registerStatus.equals("idle")) {
			registerStatus = "registering";
			MonitorLinkRegisterer monitorlinkRegisterer = new MonitorLinkRegisterer(list);
			monitorlinkRegisterer.addObserver(this);
			error = "registering";
		}
		else if (registerStatus.equals("registered")) {
			error = registerError;
			registerStatus = "idle";
			registerError = "";
		}
		
		error = "<return>" + error + "</return>";
		
		return error;
	}
	
	/**
	 * Does execute calls
	 * @param uriList list of control system addresses (Format: &#060;uri&#062;protocol&#058;&#047;&#047;address&#060;&#047;uri&#062;...&#060;uri&#062;protocol&#058;&#047;&#047;address&#060;&#047;uri&#062;)
	 * @return return data
	 */
	public String doCdaExecute(String uriList) {
		
		String[] str = null;
		String ret = "<return></return>";
		
		if (userPrivilege == supervisor || userPrivilege == expert) {
			str = uriList.split("</uri>");
			for (int i = 0; i < str.length; i++) {
				commonDeviceAccess.execute(str[i].replace("<uri>", ""), true);
			}
		}
		
		return ret;
	}
	
	/**
	 * Returns the video receivers attached to this page
	 * @return list of receivers
	 */
	public ArrayList<String> getVideoReceivers() {
		
		return null;
	}
	
	/**
	 * Returns common device access instance attached to this page 
	 * @return common device access
	 */
	public Web2cCommonDeviceAccess getCommonDeviceAccess() {
		
		return commonDeviceAccess;
	}
	
	/**
	 * Returns the id of the chat channel instance attached to this page 
	 * @return id
	 */
	public String getChatChannelId() {
		
		return null;
	}
	
	/**
	 * Constructs the page
	 */
	private void constructPage() {

		Web2cConfigReader configReader = null;
		
		try {
			configReader = new Web2cConfigReader(new File(homeDir + File.separator + pageConfigFile), debugLogger, debugListener);
			decodeAttributes(configReader.getPageAttributeNames(), configReader.getPageAttributeValues());
		}
		catch (Exception e) {
			error = error + configReader.getError();
		}
	}
	
	/**
	 * Returns page construction error
	 * @return error string
	 */
	public String getError() {
		
		return error;
	}

	/**
	 * Decodes the page's attributes and sets default values.
	 */
	private void decodeAttributes(ArrayList<String> attributeNames, ArrayList<String> attributeValues) {
		
		ArrayList<String> expertName = new ArrayList<String>();
		ArrayList<String> supervisorName = new ArrayList<String>();
		ArrayList<String> expertAddress = new ArrayList<String>();
		ArrayList<String> exclusiveName = new ArrayList<String>();
		
		for (int i = 0; i < attributeNames.toArray().length; i++) {
			String name = attributeNames.get(i);
			String value = attributeValues.get(i);
			if (name.equals("exclusive")) {
				if (!value.equals("") && !value.equals(" ")) {
					exclusiveName.add(value);
				}
			}	
			if (name.equals("expert")) {
				if (!value.equals("") && !value.equals(" ")) {
					expertName.add(value);
				}
			}	
			if (name.equals("supervisor")) {
				if (!value.equals("") && !value.equals(" ")) {
					supervisorName.add(value); 
				}
			}	
			if (name.equals("ipAddress")) {
				if (!value.equals("") && !value.equals(" ")) {
					expertAddress.add(value);
				}
			}	
		}
		
		exclusive = false;
		if (exclusiveName.isEmpty()) {
			exclusive = false;
		}
		else {
			exclusive = true;
			for (Iterator<String> it = exclusiveName.iterator(); it.hasNext();) {
				String item = it.next().toString();
				if (item.equals(userName)) {
					exclusive = false;
				}
			}
		}

		networkPrivilege = layman;
		if (expertAddress.isEmpty()) {
			networkPrivilege = layman;
		}
		else {
			for (Iterator<String> it = expertAddress.iterator(); it.hasNext();) {
				String nextAddress = it.next().toString();
				if (nextAddress.replace(".", ">").split(">")[3].equals("*") || nextAddress.replace(".", ">").split(">")[3].equals(userAddress.replace(".", ">").split(">")[3])) {
					if (nextAddress.replace(".", ">").split(">")[2].equals("*") || nextAddress.replace(".", ">").split(">")[2].equals(userAddress.replace(".", ">").split(">")[2])) {
						if (nextAddress.replace(".", ">").split(">")[1].equals("*") || nextAddress.replace(".", ">").split(">")[1].equals(userAddress.replace(".", ">").split(">")[1])) {
							if (nextAddress.replace(".", ">").split(">")[0].equals("*") || nextAddress.replace(".", ">").split(">")[0].equals(userAddress.replace(".", ">").split(">")[0])) {
								networkPrivilege = expert;
							}
						}
					}
				}
			}
		}
		
		if (userAddress.equals("*.*.*.*")) {
			userPrivilege = supervisor;
		}
		else {
			userPrivilege = layman;
			if (expertName.isEmpty() && supervisorName.isEmpty()) {
				userPrivilege = layman;
			}
			if (!expertName.isEmpty()) {
				for (Iterator<String> it = expertName.iterator(); it.hasNext();) {
					String item = it.next().toString();;
					if (item.indexOf("(") < 0) {
						if (item.equals(userName)) {
							userPrivilege = expert;
						}
					}
					else {
						String name = item.substring(0, item.indexOf("("));
						String address = item.substring(item.indexOf("(") + 1, item.indexOf(")"));
						if (name.equals(userName)) {
							if (address.replace(".", ">").split(">")[3].equals("*") || address.replace(".", ">").split(">")[3].equals(userAddress.replace(".", ">").split(">")[3])) {
								if (address.replace(".", ">").split(">")[2].equals("*") || address.replace(".", ">").split(">")[2].equals(userAddress.replace(".", ">").split(">")[2])) {
									if (address.replace(".", ">").split(">")[1].equals("*") || address.replace(".", ">").split(">")[1].equals(userAddress.replace(".", ">").split(">")[1])) {
										if (address.replace(".", ">").split(">")[0].equals("*") || address.replace(".", ">").split(">")[0].equals(userAddress.replace(".", ">").split(">")[0])) {
											networkPrivilege = expert;
											userPrivilege = expert;
										}
										else {
											networkPrivilege = layman;
											userPrivilege = layman;
										}
									}
								}
							}
						}
					}
				}		
			}
			if (!supervisorName.isEmpty()) {
				for (Iterator<String> it = supervisorName.iterator(); it.hasNext();) {
					String item = it.next().toString();;
					if (item.indexOf("(") < 0) {
						if (item.equals(userName)) {
							userPrivilege = supervisor;
						}
					}
					else {
						String name = item.substring(0, item.indexOf("("));
						String address = item.substring(item.indexOf("(") + 1, item.indexOf(")"));
						if (name.equals(userName)) {
							if (address.replace(".", ">").split(">")[3].equals("*") || address.replace(".", ">").split(">")[3].equals(userAddress.replace(".", ">").split(">")[3])) {
								if (address.replace(".", ">").split(">")[2].equals("*") || address.replace(".", ">").split(">")[2].equals(userAddress.replace(".", ">").split(">")[2])) {
									if (address.replace(".", ">").split(">")[1].equals("*") || address.replace(".", ">").split(">")[1].equals(userAddress.replace(".", ">").split(">")[1])) {
										if (address.replace(".", ">").split(">")[0].equals("*") || address.replace(".", ">").split(">")[0].equals(userAddress.replace(".", ">").split(">")[0])) {
											networkPrivilege = expert;
											userPrivilege = supervisor;
										}
										else {
											networkPrivilege = layman;
											userPrivilege = layman;
										}
									}
								}
							}
						}
					}
				}
			}
			
			if (userPrivilege == expert && networkPrivilege == expert) {
				userPrivilege = expert;
			}
			else if (userPrivilege == supervisor && networkPrivilege == expert) {
				userPrivilege = supervisor;
			}
			else {
				userPrivilege = layman;
			}
		}	
	}
	
	/**
	 * Encodes a string (to hexadecimal representation %XX)
	 * @param value string to encode
	 * @return encoded string
	 */
	private String encodeString(String value) {
		
		String toHexString = "";
		
		char[] chars = value.toCharArray();
	    for(int i = 0; i < chars.length; i++) {
	    	if ((int)chars[i] <= 0x1f && (int)chars[i] != 0x0a) toHexString = toHexString + "%" + "20";
	    	else toHexString = toHexString + "%" + Integer.toHexString((int)chars[i]);
	    }
	    
	    return toHexString;		
	}	
	
	/**
	 * @author  R.Bacher, DESY
	 */
	public class MonitorLinkRegisterer extends Observable implements Runnable {

		private Thread registerer = null;
		private ArrayList<String> monitorLinkList = null;
		private String error = "";
		
		/**	
		 * @param list list of monitor links to be registered
		 */
		public MonitorLinkRegisterer (ArrayList<String> list) {
			
			super();
			
			monitorLinkList = list;
		
			if (registerer == null) {
				registerer = new Thread(this);
				registerer.start();
			}
		}
		
		/**
		 * Registers monitor links
		 */
		public void run() {	
			
			synchronized(monitorList) {
				synchronized(monitorLinkUpdaterList) {
					error = "";
					for (Iterator<String> it = monitorLinkList.iterator();it.hasNext();) {
						String prop = (String)(it.next());
						String linkError = "";
						if (!monitorList.contains(prop) && (prop.split(":")[0].equalsIgnoreCase("tine") || prop.split(":")[0].equalsIgnoreCase("tango") || prop.split(":")[0].equalsIgnoreCase("epics") || prop.split(":")[0].equalsIgnoreCase("doocs"))) {
							commonDeviceAccess.registerMonitor(prop);
							linkError = commonDeviceAccess.getError();
							if (linkError.equals("")) {
								monitorList.add(prop);
								monitorLinkUpdaterList.add(new MonitorLinkUpdater(prop));
							}
							error = error + linkError;
						}
					}
				}
			}
			
			setChanged();
			notifyObservers(error);
			
			registerer.interrupt();
		}
	}
	
	/**
	 * @author  R.Bacher, DESY
	 */
	public class MonitorLinkUpdater implements Runnable {

		private Thread updater = null;
		boolean updaterStarted = false;
		private String monitorLinkAddress = "";
		private String update = "";
		
		/**	
		 * @param address address of monitor link to be updated
		 */
		public MonitorLinkUpdater (String address) {
			
			super();
			
			monitorLinkAddress = address;
		
			if (updater == null) {
				updater = new Thread(this);
				updaterStarted = true;
				updater.start();
			}
		}
		
		/**
		 * Updates requested monitor link
		 */
		public void run() {	
			
			while (updaterStarted == true) {
				update = "<update><uri>" + monitorLinkAddress + "</uri>"; 
				String[] cdaData = commonDeviceAccess.getMonitorData(monitorLinkAddress).split("</value><timeStamp>");
				cdaData[0] = characterEscaper.decodeString(cdaData[0].replace("<value>", ""));
				if (cdaData[0].contains(",")) {
					String[] retData = cdaData[0].split(",");
					StringBuffer strBuf = new StringBuffer();
					for (int i = 0; i < retData.length; i++) {
						strBuf.append("<value>" + retData[i] + "</value>");
					}
					update = update + strBuf.toString() + "</update>";
				}
				else {
					update = update + "<value>" + cdaData[0] + "</value></update>";
				}
	
				try {Thread.sleep(gatewayUpdateInterval * 1000);}
				catch (InterruptedException e) {}
			}
			
			updater.interrupt();
		}
		
		/**
		 * Returns the most recent update
		 * @return update
		 */
		public String getMostRecentUpdate() {
			
			return update;
		}
		
		/**
		 * Terminates updating cache data
		 */
		public void stopUpdate() {
			
			updaterStarted = false;
		}
		
		/**
		 * Returns the updater status
		 * @return updater status
		 */
		public boolean getUpdaterStatus() {
			
			return updaterStarted;
		}
	}
}
