package de.desy.web2cToolkit.web2cUtil.web2cLifeCycleControl;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;

import de.desy.web2cToolkit.web2cGateway.Web2cGatewayPage;
import de.desy.web2cToolkit.web2cUtil.web2cLifeCycleControl.Web2cSessionHandler;
import de.desy.web2cToolkit.web2c.web2cComponents.Web2cPage;
import de.desy.web2cToolkit.web2cUtil.web2cInterfaces.IWeb2cConstants;

/**
 * @author R.Bacher, DESY
 */
public class Web2cGarbageCollector implements IWeb2cConstants {

	private long inactivityTimeout = inactivityTimeoutInterval * 1000;
	private int garbageCollectionCounter = 0;
	private String servletRuntimeEngine = "";
	private boolean sessionWatcherIsEnabled = sessionWatcherEnabled;
	private boolean garbageCollectionIsEnabled = garbageCollectionEnabled;
	
	/**
	 * Watches whether web client is still alive; if not the Web2cSession attached is removed from the session list.
	 * Removes sessions on demand after logout.
	 * @param runtimeEngine servlet
	 */
	public Web2cGarbageCollector(String runtimeEngine) {
		
		servletRuntimeEngine = runtimeEngine;	
		
		if (servletRuntimeEngine.equals("synopticDisplay")) {
			inactivityTimeout = inactivityTimeoutInterval * 1000;
		}				
		if (servletRuntimeEngine.equals("archiveViewer")) {
			inactivityTimeout = archiveViewerInactivityTimeoutInterval * 1000;
		}
		if (servletRuntimeEngine.equals("gateway")) {
			inactivityTimeout = gatewayInactivityTimeoutInterval * 1000;
		}
	}

	/**
	 * Clears the cache (Web2c only)
	 * @param cache cache
	 */
	public void clearCache(Web2cImageCache cache) {

		if (servletRuntimeEngine.equals("synopticDisplay")) {
			cache.clearChartCache();
			cache.clearVideoCache();
		}
	}

	/**
	 * Cancels dead sessions (Web2c, Web2cArchiveViewer, Web2cGateway)
	 * @param sessionHandlerList list of sessions
	 */
	public void removeDeadSessions(ArrayList<Web2cSessionHandler> sessionHandlerList) {
		
		long now = 0;
		long lastPoll = 0;
		ArrayList<Web2cSessionHandler> sessionHandlerToRemoveList = null;
		
		synchronized(sessionHandlerList) {
			if (sessionWatcherIsEnabled == true) {
				sessionHandlerToRemoveList = new ArrayList<Web2cSessionHandler>();
				for (Iterator<Web2cSessionHandler> it = sessionHandlerList.iterator(); it.hasNext();) {
					Web2cSessionHandler sessionHandler = (Web2cSessionHandler)it.next();
					if (sessionHandler != null) {		
						now = Calendar.getInstance().getTimeInMillis();
						lastPoll = Long.valueOf(sessionHandler.getAttribute("lastPoll"));
						if (lastPoll != 0) {
							if ((now - lastPoll) > inactivityTimeout) {
//								System.out.println("Web2cGarbageCollector cancels sessions (Web2c): " + sessionHandler.getAttribute("id") + " " + now + " " + lastPoll + " " + (now - lastPoll) + ">" + inactivityTimeout);
								sessionHandlerToRemoveList.add(sessionHandler);						
							}
							else {
//								System.out.println("Web2cGarbageCollector is continuing: " + sessionHandler.getAttribute("id") + " " + now + " " + lastPoll + " " + (now - lastPoll) + "<=" + inactivityTimeout);
							}
						}
					}
				}
				
				for (Iterator<Web2cSessionHandler> it = sessionHandlerToRemoveList.iterator(); it.hasNext();) {
					Web2cSessionHandler sessionHandlerToRemove = (Web2cSessionHandler)it.next();
					Object page = null;
					if (servletRuntimeEngine.equals("synopticDisplay")) {
						page = (Object)sessionHandlerToRemove.getWeb2cPage();
					}
					if (servletRuntimeEngine.equals("archiveViewer")) {
						page = (Object)sessionHandlerToRemove.getWeb2cArchiveViewerPage();
					}
					if (servletRuntimeEngine.equals("gateway")) {
						page = (Object)sessionHandlerToRemove.getWeb2cGatewayPage();
					}
//					System.out.println("Web2c session canceled: ID=" + sessionHandlerToRemove.getAttribute("id") + " ConfigFile=" + sessionHandlerToRemove.getAttribute("loadedConfigFileName") + " CancelTime=" + String.valueOf(Calendar.getInstance().getTime().toString()) + " ActiveWeb2cSessions=" + (sessionHandlerList.toArray().length - 1));		
					releaseSessionResources(page);
					sessionHandlerList.remove(sessionHandlerToRemove);
				}
			}
		}
		
		requestJavaGarbageCollection();
	}
	
	/**
	 * Removes a session (Web2c, Web2cArchiveViewer, Web2cGateway) after logout
	 * @param sessionHandlerList list of sessions
	 * @param sessionHandler handler of session to be removed
	 */
	public void removeSession(ArrayList<Web2cSessionHandler> sessionHandlerList, Web2cSessionHandler sessionHandler) {
		
		synchronized(sessionHandlerList) {
//			System.out.println("Web2c session canceled: ID=" + sessionHandler.getAttribute("id") + " ConfigFile=" + sessionHandler.getAttribute("loadedConfigFileName") + " CancelTime=" + String.valueOf(Calendar.getInstance().getTime().toString()) + " ActiveWeb2cSessions=" + (sessionHandlerList.toArray().length - 1));
			Object page = null;
			if (servletRuntimeEngine.equals("synopticDisplay")) {
				page = (Object)sessionHandler.getWeb2cPage();
			}
			if (servletRuntimeEngine.equals("archiveViewer")) {
				page = (Object)sessionHandler.getWeb2cArchiveViewerPage();
			}
			if (servletRuntimeEngine.equals("gateway")) {
				page = (Object)sessionHandler.getWeb2cGatewayPage();
			}
			releaseSessionResources(page);
			sessionHandlerList.remove(sessionHandler);
		}
		
		requestJavaGarbageCollection();
	}
	
	/**
	 * Releases the resources (chat, video, device access) attached to the session (page) to be removed
	 * @param page page 
	 */
	private void releaseSessionResources(Object page) {
		
		if (page != null) {
			if (servletRuntimeEngine.equals("synopticDisplay")) {
				((Web2cPage)page).stopUpdate();
				
				ArrayList<String> monitors = ((Web2cPage)page).getMonitors();		
				if (monitors != null && !monitors.isEmpty()) {
					for (Iterator<String> it = monitors.iterator(); it.hasNext();) {
						String monitorId = (String)it.next();
						((Web2cPage)page).getCommonDeviceAccess().removeRequester(monitorId);
					}
				}
				
				ArrayList<String> receivers = ((Web2cPage)page).getVideoReceivers();
				if (receivers != null && !receivers.isEmpty()) {
					for (Iterator<String> it = receivers.iterator(); it.hasNext();) {
						String receiverId = (String)it.next();
						((Web2cPage)page).getCommonVideoConnector().removeRequester(receiverId);
					}
				}
				
				String chatChannelId = ((Web2cPage)page).getChatChannelId();
				if (chatChannelId != null) {
					((Web2cPage)page).getChatHandler().removeRequester(chatChannelId);
				}
			}
			
			if (servletRuntimeEngine.equals("gateway")) {
				((Web2cGatewayPage)page).stopUpdate();
				
				ArrayList<String> monitors = ((Web2cGatewayPage)page).getMonitors();		
				if (monitors != null && !monitors.isEmpty()) {
					for (Iterator<String> it = monitors.iterator(); it.hasNext();) {
						String monitorId = (String)it.next();
						((Web2cGatewayPage)page).getCommonDeviceAccess().removeRequester(monitorId);
					}
				}
			}
		}
	}
	
	/**
	 * Requests Java garbage collection
	 */
	private void requestJavaGarbageCollection() {
		
		if (garbageCollectionIsEnabled == true) {
			garbageCollectionCounter = garbageCollectionCounter + 1;
			if (garbageCollectionCounter > maxGarbageCollectionCounter) {
				garbageCollectionCounter = 0;
				System.gc();
//				System.out.println("Web2cGarbageCollector requests JAVA Garbage Collector");
			}
		}
	}
}
